from __future__ import annotations

import datetime as dt
from contextlib import suppress
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from re import DOTALL
from types import NoneType
from typing import TYPE_CHECKING, Any, override

from utilities.datetime import (
    is_instance_date_not_datetime,
    is_subclass_date_not_datetime,
)
from utilities.enum import ParseEnumError, parse_enum
from utilities.functions import is_subclass_not_bool_int
from utilities.iterables import OneEmptyError, OneNonUniqueError, one, one_str
from utilities.math import ParseNumberError, parse_number
from utilities.re import ExtractGroupError, extract_group
from utilities.sentinel import ParseSentinelError, Sentinel, parse_sentinel
from utilities.text import (
    BRACKETS,
    LIST_SEPARATOR,
    PAIR_SEPARATOR,
    ParseBoolError,
    ParseNoneError,
    join_strs,
    parse_bool,
    parse_none,
    split_key_value_pairs,
    split_str,
)
from utilities.types import Duration, Number, ParseObjectExtra, SerializeObjectExtra
from utilities.typing import (
    get_args,
    is_dict_type,
    is_frozenset_type,
    is_list_type,
    is_literal_type,
    is_optional_type,
    is_set_type,
    is_tuple_type,
    is_union_type,
)
from utilities.version import ParseVersionError, Version, parse_version

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from collections.abc import Set as AbstractSet


def parse_object(
    type_: Any,
    text: str,
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    head: bool = False,
    case_sensitive: bool = False,
    extra: ParseObjectExtra | None = None,
) -> Any:
    """Parse text."""
    if type_ is None:
        try:
            return parse_none(text)
        except ParseNoneError:
            raise _ParseObjectParseError(type_=type_, text=text) from None
    if isinstance(type_, type):
        return _parse_object_type(
            type_, text, case_sensitive=case_sensitive, extra=extra
        )
    if is_dict_type(type_):
        return _parse_object_dict_type(
            type_,
            text,
            list_separator=list_separator,
            pair_separator=pair_separator,
            brackets=brackets,
            head=head,
            case_sensitive=case_sensitive,
            extra=extra,
        )
    if is_frozenset_type(type_):
        return frozenset(
            _parse_object_set_type(
                type_,
                text,
                list_separator=list_separator,
                pair_separator=pair_separator,
                brackets=brackets,
                head=head,
                case_sensitive=case_sensitive,
                extra=extra,
            )
        )
    if is_list_type(type_):
        return _parse_object_list_type(
            type_,
            text,
            list_separator=list_separator,
            pair_separator=pair_separator,
            brackets=brackets,
            head=head,
            case_sensitive=case_sensitive,
            extra=extra,
        )
    if is_literal_type(type_):
        return one_str(get_args(type_), text, head=head, case_sensitive=case_sensitive)
    if is_optional_type(type_):
        with suppress(ParseNoneError):
            return parse_none(text)
        inner = one(arg for arg in get_args(type_) if arg is not NoneType)
        try:
            return parse_object(
                inner,
                text,
                list_separator=list_separator,
                pair_separator=pair_separator,
                brackets=brackets,
                head=head,
                case_sensitive=case_sensitive,
                extra=extra,
            )
        except _ParseObjectParseError:
            raise _ParseObjectParseError(type_=type_, text=text) from None
    if is_set_type(type_):
        return _parse_object_set_type(
            type_,
            text,
            list_separator=list_separator,
            pair_separator=pair_separator,
            brackets=brackets,
            head=head,
            case_sensitive=case_sensitive,
            extra=extra,
        )
    if is_tuple_type(type_):
        return _parse_object_tuple_type(
            type_,
            text,
            list_separator=list_separator,
            pair_separator=pair_separator,
            brackets=brackets,
            head=head,
            case_sensitive=case_sensitive,
            extra=extra,
        )
    if is_union_type(type_):
        return _parse_object_union_type(type_, text, extra=extra)
    raise _ParseObjectParseError(type_=type_, text=text) from None


def _parse_object_type(
    cls: type[Any],
    text: str,
    /,
    *,
    case_sensitive: bool = False,
    extra: ParseObjectExtra | None = None,
) -> Any:
    """Parse text."""
    if issubclass(cls, NoneType):
        try:
            return parse_none(text)
        except ParseNoneError:
            raise _ParseObjectParseError(type_=cls, text=text) from None
    if issubclass(cls, str):
        return text
    if issubclass(cls, bool):
        try:
            return parse_bool(text)
        except ParseBoolError:
            raise _ParseObjectParseError(type_=cls, text=text) from None
    if is_subclass_not_bool_int(cls, int):
        try:
            return int(text)
        except ValueError:
            raise _ParseObjectParseError(type_=cls, text=text) from None
    if issubclass(cls, float):
        try:
            return float(text)
        except ValueError:
            raise _ParseObjectParseError(type_=cls, text=text) from None
    if issubclass(cls, Enum):
        try:
            return parse_enum(text, cls, case_sensitive=case_sensitive)
        except ParseEnumError:
            raise _ParseObjectParseError(type_=cls, text=text) from None
    if issubclass(cls, Path):
        return Path(text).expanduser()
    if issubclass(cls, Sentinel):
        try:
            return parse_sentinel(text)
        except ParseSentinelError:
            raise _ParseObjectParseError(type_=cls, text=text) from None
    if issubclass(cls, Version):
        try:
            return parse_version(text)
        except ParseVersionError:
            raise _ParseObjectParseError(type_=cls, text=text) from None
    if is_subclass_date_not_datetime(cls):
        from utilities.whenever import ParseDateError, parse_date

        try:
            return parse_date(text)
        except ParseDateError:
            raise _ParseObjectParseError(type_=cls, text=text) from None
    if issubclass(cls, dt.datetime):
        from utilities.whenever import ParseDateTimeError, parse_datetime

        try:
            return parse_datetime(text)
        except ParseDateTimeError:
            raise _ParseObjectParseError(type_=cls, text=text) from None
    if issubclass(cls, dt.time):
        from utilities.whenever import ParseTimeError, parse_time

        try:
            return parse_time(text)
        except ParseTimeError:
            raise _ParseObjectParseError(type_=cls, text=text) from None
    if issubclass(cls, dt.timedelta):
        from utilities.whenever import ParseTimedeltaError, parse_timedelta

        try:
            return parse_timedelta(text)
        except ParseTimedeltaError:
            raise _ParseObjectParseError(type_=cls, text=text) from None
    if extra is not None:
        try:
            parser = one(p for c, p in extra.items() if issubclass(cls, c))
        except OneEmptyError:
            pass
        except OneNonUniqueError as error:
            raise _ParseObjectExtraNonUniqueError(
                type_=cls, text=text, first=error.first, second=error.second
            ) from None
        else:
            return parser(text)
    raise _ParseObjectParseError(type_=cls, text=text)


def _parse_object_dict_type(
    type_: Any,
    text: str,
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    head: bool = False,
    case_sensitive: bool = False,
    extra: ParseObjectExtra | None = None,
) -> dict[Any, Any]:
    key_type, value_type = get_args(type_)
    try:
        inner_text = extract_group(r"^{(.*)}$", text, flags=DOTALL)
    except ExtractGroupError:
        raise _ParseObjectParseError(type_=type_, text=text) from None
    pairs = split_key_value_pairs(
        inner_text,
        list_separator=list_separator,
        pair_separator=pair_separator,
        brackets=brackets,
        mapping=True,
    )
    keys = (
        parse_object(
            key_type,
            k,
            list_separator=list_separator,
            pair_separator=pair_separator,
            brackets=brackets,
            head=head,
            case_sensitive=case_sensitive,
            extra=extra,
        )
        for k in pairs
    )
    values = (
        parse_object(
            value_type,
            v,
            list_separator=list_separator,
            pair_separator=pair_separator,
            brackets=brackets,
            head=head,
            case_sensitive=case_sensitive,
            extra=extra,
        )
        for v in pairs.values()
    )
    try:
        return dict(zip(keys, values, strict=True))
    except _ParseObjectParseError:
        raise _ParseObjectParseError(type_=type_, text=text) from None


def _parse_object_list_type(
    type_: Any,
    text: str,
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    head: bool = False,
    case_sensitive: bool = False,
    extra: ParseObjectExtra | None = None,
) -> list[Any]:
    inner_type = one(get_args(type_))
    try:
        inner_text = extract_group(r"^\[(.*)\]$", text, flags=DOTALL)
    except ExtractGroupError:
        raise _ParseObjectParseError(type_=type_, text=text) from None
    texts = split_str(inner_text, separator=list_separator)
    try:
        return [
            parse_object(
                inner_type,
                t,
                list_separator=list_separator,
                pair_separator=pair_separator,
                brackets=brackets,
                head=head,
                case_sensitive=case_sensitive,
                extra=extra,
            )
            for t in texts
        ]
    except _ParseObjectParseError:
        raise _ParseObjectParseError(type_=type_, text=text) from None


def _parse_object_set_type(
    type_: Any,
    text: str,
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    head: bool = False,
    case_sensitive: bool = False,
    extra: ParseObjectExtra | None = None,
) -> set[Any]:
    inner_type = one(get_args(type_))
    try:
        inner_text = extract_group(r"^{(.*)}$", text, flags=DOTALL)
    except ExtractGroupError:
        raise _ParseObjectParseError(type_=type_, text=text) from None
    texts = split_str(inner_text, separator=list_separator)
    try:
        return {
            parse_object(
                inner_type,
                t,
                list_separator=list_separator,
                pair_separator=pair_separator,
                brackets=brackets,
                head=head,
                case_sensitive=case_sensitive,
                extra=extra,
            )
            for t in texts
        }
    except _ParseObjectParseError:
        raise _ParseObjectParseError(type_=type_, text=text) from None


def _parse_object_union_type(
    type_: Any, text: str, /, *, extra: ParseObjectExtra | None = None
) -> Any:
    if type_ is Number:
        try:
            return parse_number(text)
        except ParseNumberError:
            raise _ParseObjectParseError(type_=type_, text=text) from None
    if type_ is Duration:
        from utilities.whenever import ParseDurationError, parse_duration

        try:
            return parse_duration(text)
        except ParseDurationError:
            raise _ParseObjectParseError(type_=type_, text=text) from None
    if extra is not None:
        try:
            parser = one(p for c, p in extra.items() if c is type_)
        except OneEmptyError:
            pass
        else:
            return parser(text)
    raise _ParseObjectParseError(type_=type_, text=text) from None


def _parse_object_tuple_type(
    type_: Any,
    text: str,
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    head: bool = False,
    case_sensitive: bool = False,
    extra: ParseObjectExtra | None = None,
) -> tuple[Any, ...]:
    args = get_args(type_)
    try:
        inner = extract_group(r"^\((.*)\)$", text, flags=DOTALL)
    except ExtractGroupError:
        raise _ParseObjectParseError(type_=type_, text=text) from None
    texts = inner.split(list_separator)
    if len(args) != len(texts):
        raise _ParseObjectParseError(type_=type_, text=text)
    try:
        return tuple(
            parse_object(
                arg,
                text,
                list_separator=list_separator,
                pair_separator=pair_separator,
                brackets=brackets,
                head=head,
                case_sensitive=case_sensitive,
                extra=extra,
            )
            for arg, text in zip(args, texts, strict=True)
        )
    except _ParseObjectParseError:
        raise _ParseObjectParseError(type_=type_, text=text) from None


@dataclass
class ParseObjectError(Exception):
    type_: Any
    text: str


@dataclass
class _ParseObjectParseError(ParseObjectError):
    @override
    def __str__(self) -> str:
        return f"Unable to parse {self.type_!r}; got {self.text!r}"


@dataclass
class _ParseObjectExtraNonUniqueError(ParseObjectError):
    first: type[Any]
    second: type[Any]

    @override
    def __str__(self) -> str:
        return f"Unable to parse {self.type_!r} since `extra` must contain exactly one parent class; got {self.first!r}, {self.second!r} and perhaps more"


##


def serialize_object(
    obj: Any,
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    extra: SerializeObjectExtra | None = None,
) -> str:
    """Convert an object to text."""
    if (obj is None) or isinstance(
        obj, bool | int | float | str | Path | Sentinel | Version
    ):
        return str(obj)
    if is_instance_date_not_datetime(obj):
        from utilities.whenever import serialize_date

        return serialize_date(obj)
    if isinstance(obj, dt.datetime):
        from utilities.whenever import serialize_datetime

        return serialize_datetime(obj)
    if isinstance(obj, dt.time):
        from utilities.whenever import serialize_time

        return serialize_time(obj)
    if isinstance(obj, dt.timedelta):
        from utilities.whenever import serialize_timedelta

        return serialize_timedelta(obj)
    if isinstance(obj, Enum):
        return obj.name
    if isinstance(obj, dict):
        return _serialize_object_dict(
            obj, list_separator=list_separator, pair_separator=pair_separator
        )
    if isinstance(obj, list):
        return _serialize_object_list(
            obj, list_separator=list_separator, pair_separator=pair_separator
        )
    if isinstance(obj, tuple):
        return _serialize_object_tuple(
            obj, list_separator=list_separator, pair_separator=pair_separator
        )
    if isinstance(obj, set | frozenset):
        return _serialize_object_set(
            obj, list_separator=list_separator, pair_separator=pair_separator
        )
    if extra is not None:
        return _serialize_object_extra(obj, extra)
    raise _SerializeObjectSerializeError(obj=obj)


def _serialize_object_dict(
    obj: Mapping[Any, Any],
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
) -> str:
    keys = (
        serialize_object(
            k, list_separator=list_separator, pair_separator=pair_separator
        )
        for k in obj
    )
    values = (
        serialize_object(
            v, list_separator=list_separator, pair_separator=pair_separator
        )
        for v in obj.values()
    )
    items = zip(keys, values, strict=True)
    joined_items = (join_strs(item, separator=pair_separator) for item in items)
    joined = join_strs(joined_items, separator=list_separator)
    return f"{{{joined}}}"


def _serialize_object_extra(obj: Any, extra: SerializeObjectExtra, /) -> str:
    try:
        serializer = one(
            s
            for c, s in extra.items()
            if (isinstance(c, type) and isinstance(obj, c))
            or isinstance(obj, get_args(c))
        )
    except OneEmptyError:
        raise _SerializeObjectSerializeError(obj=obj) from None
    except OneNonUniqueError as error:
        raise _SerializeObjectExtraNonUniqueError(
            obj=obj, first=error.first, second=error.second
        ) from None
    else:
        return serializer(obj)


def _serialize_object_list(
    obj: Sequence[Any],
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
) -> str:
    items = (
        serialize_object(
            i, list_separator=list_separator, pair_separator=pair_separator
        )
        for i in obj
    )
    joined = join_strs(items, separator=list_separator)
    return f"[{joined}]"


def _serialize_object_set(
    obj: AbstractSet[Any],
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
) -> str:
    items = (
        serialize_object(
            i, list_separator=list_separator, pair_separator=pair_separator
        )
        for i in obj
    )
    joined = join_strs(items, sort=True, separator=list_separator)
    return f"{{{joined}}}"


def _serialize_object_tuple(
    obj: tuple[Any, ...],
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
) -> str:
    items = (
        serialize_object(
            i, list_separator=list_separator, pair_separator=pair_separator
        )
        for i in obj
    )
    joined = join_strs(items, separator=list_separator)
    return f"({joined})"


@dataclass(kw_only=True, slots=True)
class SerializeObjectError(Exception):
    obj: Any


class _SerializeObjectSerializeError(SerializeObjectError):
    @override
    def __str__(self) -> str:
        return f"Unable to serialize object {self.obj!r}"


@dataclass
class _SerializeObjectExtraNonUniqueError(SerializeObjectError):
    first: type[Any]
    second: type[Any]

    @override
    def __str__(self) -> str:
        return f"Unable to serialize object {self.obj!r} since `extra` must contain exactly one parent class; got {self.first!r}, {self.second!r} and perhaps more"


__all__ = ["parse_object", "serialize_object"]



from __future__ import annotations

from types import NoneType

from utilities.ipython import check_ipython_class, is_ipython


class TestCheckIPythonClass:
    def test_main(self) -> None:
        assert not check_ipython_class(NoneType)


class TestIsIPython:
    def test_main(self) -> None:
        assert not is_ipython()



from __future__ import annotations

import datetime as dt
import re
from contextlib import contextmanager
from dataclasses import dataclass, field
from functools import cached_property
from itertools import product
from logging import (
    DEBUG,
    ERROR,
    NOTSET,
    FileHandler,
    Formatter,
    Handler,
    Logger,
    LogRecord,
    StreamHandler,
    basicConfig,
    getLevelNamesMapping,
    getLogger,
    setLogRecordFactory,
)
from logging.handlers import BaseRotatingHandler, TimedRotatingFileHandler
from pathlib import Path
from re import Pattern, search
from sys import stdout
from time import time
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Literal,
    Self,
    assert_never,
    cast,
    override,
)

from utilities.atomicwrites import move_many, writer
from utilities.dataclasses import replace_non_sentinel
from utilities.datetime import (
    SECOND,
    get_now_local,
    maybe_sub_pct_y,
    parse_datetime_compact,
    round_datetime,
    serialize_compact,
)
from utilities.errors import ImpossibleCaseError
from utilities.git import get_repo_root
from utilities.iterables import OneEmptyError, always_iterable, one
from utilities.pathlib import ensure_suffix, resolve_path
from utilities.reprlib import (
    RICH_EXPAND_ALL,
    RICH_INDENT_SIZE,
    RICH_MAX_DEPTH,
    RICH_MAX_LENGTH,
    RICH_MAX_STRING,
    RICH_MAX_WIDTH,
)
from utilities.sentinel import Sentinel, sentinel
from utilities.traceback import RichTracebackFormatter

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator
    from logging import _FilterType
    from zoneinfo import ZoneInfo

    from utilities.types import (
        LoggerOrName,
        LogLevel,
        MaybeIterable,
        PathLike,
        PathLikeOrCallable,
    )
    from utilities.version import MaybeCallableVersionLike

try:
    from whenever import ZonedDateTime
except ModuleNotFoundError:  # pragma: no cover
    ZonedDateTime = None


##


type _When = Literal[
    "S", "M", "H", "D", "midnight", "W0", "W1", "W2", "W3", "W4", "W5", "W6"
]


class SizeAndTimeRotatingFileHandler(BaseRotatingHandler):
    """Handler which rotates on size & time."""

    stream: Any

    @override
    def __init__(
        self,
        filename: PathLike,
        mode: Literal["a", "w", "x"] = "a",
        encoding: str | None = None,
        delay: bool = False,
        errors: Literal["strict", "ignore", "replace"] | None = None,
        maxBytes: int = 0,
        when: _When = "midnight",
        interval: int = 1,
        backupCount: int = 0,
        utc: bool = False,
        atTime: dt.time | None = None,
    ) -> None:
        filename = str(Path(filename))
        super().__init__(filename, mode, encoding=encoding, delay=delay, errors=errors)
        self._max_bytes = maxBytes if maxBytes >= 1 else None
        self._backup_count = backupCount if backupCount >= 1 else None
        self._filename = Path(self.baseFilename)
        self._directory = self._filename.parent
        self._stem = self._filename.stem
        self._suffix = self._filename.suffix
        self._patterns = _compute_rollover_patterns(self._stem, self._suffix)
        self._time_handler = TimedRotatingFileHandler(
            filename,
            when=when,
            interval=interval,
            backupCount=backupCount,
            encoding=encoding,
            delay=delay,
            utc=utc,
            atTime=atTime,
            errors=errors,
        )

    @override
    def emit(self, record: LogRecord) -> None:
        try:  # skipif-ci-and-windows
            if (self._backup_count is not None) and self._should_rollover(record):
                self._do_rollover(backup_count=self._backup_count)
            FileHandler.emit(self, record)
        except Exception:  # noqa: BLE001  # pragma: no cover
            self.handleError(record)

    def _do_rollover(self, *, backup_count: int = 1) -> None:
        if self.stream:  # pragma: no cover
            self.stream.close()
            self.stream = None

        actions = _compute_rollover_actions(  # skipif-ci-and-windows
            self._directory,
            self._stem,
            self._suffix,
            patterns=self._patterns,
            backup_count=backup_count,
        )
        actions.do()  # skipif-ci-and-windows

        if not self.delay:  # pragma: no cover
            self.stream = self._open()
        self._time_handler.rolloverAt = (  # skipif-ci-and-windows
            self._time_handler.computeRollover(int(time()))
        )

    def _should_rollover(self, record: LogRecord, /) -> bool:
        if self._max_bytes is not None:  # skipif-ci-and-windows
            try:
                current = self._filename.stat().st_size
            except FileNotFoundError:
                pass
            else:
                delta = len(f"{self.format(record)}\n")
                new = current + delta
                if new >= self._max_bytes:
                    return True
        return bool(self._time_handler.shouldRollover(record))  # skipif-ci-and-windows


def _compute_rollover_patterns(stem: str, suffix: str, /) -> _RolloverPatterns:
    return _RolloverPatterns(
        pattern1=re.compile(rf"^{stem}\.(\d+){suffix}$"),
        pattern2=re.compile(rf"^{stem}\.(\d+)__(\d{{8}}T\d{{6}}){suffix}$"),
        pattern3=re.compile(
            rf"^{stem}\.(\d+)__(\d{{8}}T\d{{6}})__(\d{{8}}T\d{{6}}){suffix}$"
        ),
    )


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class _RolloverPatterns:
    pattern1: Pattern[str]
    pattern2: Pattern[str]
    pattern3: Pattern[str]


def _compute_rollover_actions(
    directory: Path,
    stem: str,
    suffix: str,
    /,
    *,
    patterns: _RolloverPatterns | None = None,
    backup_count: int = 1,
) -> _RolloverActions:
    patterns = (  # skipif-ci-and-windows
        _compute_rollover_patterns(stem, suffix) if patterns is None else patterns
    )
    files = {  # skipif-ci-and-windows
        file
        for path in directory.iterdir()
        if (file := _RotatingLogFile.from_path(path, stem, suffix, patterns=patterns))
        is not None
    }
    deletions: set[_Deletion] = set()  # skipif-ci-and-windows
    rotations: set[_Rotation] = set()  # skipif-ci-and-windows
    for file in files:  # skipif-ci-and-windows
        match file.index, file.start, file.end:
            case int() as index, _, _ if index >= backup_count:
                deletions.add(_Deletion(file=file))
            case index, None, _:
                if index is None:
                    curr = 0
                    end = get_now_local()
                else:
                    curr = index
                    end = sentinel
                try:
                    start = one(f for f in files if f.index == curr + 1).end
                except OneEmptyError:
                    start = None
                rotations.add(
                    _Rotation(file=file, index=curr + 1, start=start, end=end)
                )
            case int() as index, dt.datetime(), dt.datetime():
                rotations.add(_Rotation(file=file, index=index + 1))
            case _:  # pragma: no cover
                raise NotImplementedError
    return _RolloverActions(  # skipif-ci-and-windows
        deletions=deletions, rotations=rotations
    )


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class _RolloverActions:
    deletions: set[_Deletion] = field(default_factory=set)
    rotations: set[_Rotation] = field(default_factory=set)

    def do(self) -> None:
        for deletion in self.deletions:  # skipif-ci-and-windows
            deletion.delete()
        move_many(  # skipif-ci-and-windows
            *((r.file.path, r.destination) for r in self.rotations)
        )


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class _RotatingLogFile:
    directory: Path
    stem: str
    suffix: str
    index: int | None = None
    start: dt.datetime | None = None
    end: dt.datetime | None = None

    def __post_init__(self) -> None:
        if self.start is not None:
            self.start = round_datetime(self.start, SECOND)
        if self.end is not None:
            self.end = round_datetime(self.end, SECOND)

    @classmethod
    def from_path(
        cls,
        path: Path,
        stem: str,
        suffix: str,
        /,
        *,
        patterns: _RolloverPatterns | None = None,
    ) -> Self | None:
        if (not path.stem.startswith(stem)) or path.suffix != suffix:
            return None
        if patterns is None:  # skipif-ci-and-windows
            patterns = _compute_rollover_patterns(stem, suffix)
        try:
            (index,) = patterns.pattern1.findall(path.name)
        except ValueError:
            pass
        else:
            return cls(
                directory=path.parent, stem=stem, suffix=suffix, index=int(index)
            )
        try:
            ((index, end),) = patterns.pattern2.findall(path.name)
        except ValueError:
            pass
        else:
            return cls(
                directory=path.parent,
                stem=stem,
                suffix=suffix,
                index=int(index),
                end=parse_datetime_compact(end),
            )
        try:
            ((index, start, end),) = patterns.pattern3.findall(path.name)
        except ValueError:
            return cls(directory=path.parent, stem=stem, suffix=suffix)
        else:
            return cls(
                directory=path.parent,
                stem=stem,
                suffix=suffix,
                index=int(index),
                start=parse_datetime_compact(start),
                end=parse_datetime_compact(end),
            )

    @cached_property
    def path(self) -> Path:
        """The full path."""
        match self.index, self.start, self.end:
            case None, None, None:
                tail = None
            case int() as index, None, None:
                tail = str(index)
            case int() as index, None, dt.datetime() as end:
                tail = f"{index}__{serialize_compact(end)}"
            case int() as index, dt.datetime() as start, dt.datetime() as end:
                tail = f"{index}__{serialize_compact(start)}__{serialize_compact(end)}"
            case _:  # pragma: no cover
                raise ImpossibleCaseError(
                    case=[f"{self.index=}", f"{self.start=}", f"{self.end=}"]
                )
        stem = self.stem if tail is None else f"{self.stem}.{tail}"
        return ensure_suffix(self.directory.joinpath(stem), self.suffix)

    def replace(
        self,
        *,
        index: int | None | Sentinel = sentinel,
        start: dt.datetime | None | Sentinel = sentinel,
        end: dt.datetime | None | Sentinel = sentinel,
    ) -> Self:
        return replace_non_sentinel(  # skipif-ci-and-windows
            self, index=index, start=start, end=end
        )


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class _Deletion:
    file: _RotatingLogFile

    def delete(self) -> None:
        self.file.path.unlink(missing_ok=True)  # skipif-ci-and-windows


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class _Rotation:
    file: _RotatingLogFile
    index: int = 0
    start: dt.datetime | None | Sentinel = sentinel
    end: dt.datetime | Sentinel = sentinel

    def __post_init__(self) -> None:
        if isinstance(self.start, dt.datetime):  # skipif-ci-and-windows
            self.start = round_datetime(self.start, SECOND)
        if isinstance(self.end, dt.datetime):  # skipif-ci-and-windows
            self.end = round_datetime(self.end, SECOND)

    @cached_property
    def destination(self) -> Path:
        return self.file.replace(  # skipif-ci-and-windows
            index=self.index, start=self.start, end=self.end
        ).path


##


class StandaloneFileHandler(Handler):
    """Handler for emitting tracebacks to individual files."""

    @override
    def __init__(
        self, *, level: int = NOTSET, path: PathLikeOrCallable | None = None
    ) -> None:
        super().__init__(level=level)
        self._path = path

    @override
    def emit(self, record: LogRecord) -> None:
        try:
            path = (
                resolve_path(path=self._path)
                .joinpath(serialize_compact(get_now_local()))
                .with_suffix(".txt")
            )
            formatted = self.format(record)
            with writer(path, overwrite=True) as temp, temp.open(mode="w") as fh:
                _ = fh.write(formatted)
        except Exception:  # noqa: BLE001  # pragma: no cover
            self.handleError(record)


##


def add_filters(handler: Handler, /, *filters: _FilterType) -> None:
    """Add a set of filters to a handler."""
    for filter_i in filters:
        handler.addFilter(filter_i)


##


def basic_config(
    *,
    format: str = "{asctime} | {name} | {levelname:8} | {message}",  # noqa: A002
    level: LogLevel = "INFO",
) -> None:
    """Do the basic config."""
    basicConfig(
        format=format,
        datefmt=maybe_sub_pct_y("%Y-%m-%d %H:%M:%S"),
        style="{",
        level=level,
    )


##


def filter_for_key(
    key: str, /, *, default: bool = False
) -> Callable[[LogRecord], bool]:
    """Make a filter for a given attribute."""
    if (key in _FILTER_FOR_KEY_BLACKLIST) or key.startswith("__"):
        raise FilterForKeyError(key=key)

    def filter_(record: LogRecord, /) -> bool:
        try:
            value = getattr(record, key)
        except AttributeError:
            return default
        return bool(value)

    return filter_


# fmt: off
_FILTER_FOR_KEY_BLACKLIST = {
    "args", "created", "exc_info", "exc_text", "filename", "funcName", "getMessage", "levelname", "levelno", "lineno", "module", "msecs", "msg", "name", "pathname", "process", "processName", "relativeCreated", "stack_info", "taskName", "thread", "threadName"
}
# fmt: on


@dataclass(kw_only=True, slots=True)
class FilterForKeyError(Exception):
    key: str

    @override
    def __str__(self) -> str:
        return f"Invalid key: {self.key!r}"


##


def get_default_logging_path() -> Path:
    """Get the logging default path."""
    return get_repo_root().joinpath(".logs")


##


def get_logger(*, logger: LoggerOrName | None = None) -> Logger:
    """Get a logger."""
    match logger:
        case Logger():
            return logger
        case str() | None:
            return getLogger(logger)
        case _ as never:
            assert_never(never)


##


def get_logging_level_number(level: LogLevel, /) -> int:
    """Get the logging level number."""
    mapping = getLevelNamesMapping()
    try:
        return mapping[level]
    except KeyError:
        raise GetLoggingLevelNumberError(level=level) from None


@dataclass(kw_only=True, slots=True)
class GetLoggingLevelNumberError(Exception):
    level: LogLevel

    @override
    def __str__(self) -> str:
        return f"Invalid logging level: {self.level!r}"


##


def setup_logging(
    *,
    logger: LoggerOrName | None = None,
    console_level: LogLevel | None = "INFO",
    console_filters: Iterable[_FilterType] | None = None,
    console_fmt: str = "❯ {_zoned_datetime_str} | {name}:{funcName}:{lineno} | {message}",  # noqa: RUF001
    files_dir: PathLikeOrCallable | None = get_default_logging_path,
    files_when: _When = "D",
    files_interval: int = 1,
    files_backup_count: int = 10,
    files_max_bytes: int = 10 * 1024**2,
    files_filters: Iterable[_FilterType] | None = None,
    files_fmt: str = "{_zoned_datetime_str} | {name}:{funcName}:{lineno} | {levelname:8} | {message}",
    filters: MaybeIterable[_FilterType] | None = None,
    formatter_version: MaybeCallableVersionLike | None = None,
    formatter_max_width: int = RICH_MAX_WIDTH,
    formatter_indent_size: int = RICH_INDENT_SIZE,
    formatter_max_length: int | None = RICH_MAX_LENGTH,
    formatter_max_string: int | None = RICH_MAX_STRING,
    formatter_max_depth: int | None = RICH_MAX_DEPTH,
    formatter_expand_all: bool = RICH_EXPAND_ALL,
    extra: Callable[[LoggerOrName | None], None] | None = None,
) -> None:
    """Set up logger."""
    # log record factory
    from utilities.tzlocal import get_local_time_zone  # skipif-ci-and-windows

    class LogRecordNanoLocal(  # skipif-ci-and-windows
        _AdvancedLogRecord, time_zone=get_local_time_zone()
    ): ...

    setLogRecordFactory(LogRecordNanoLocal)  # skipif-ci-and-windows

    console_fmt, files_fmt = [  # skipif-ci-and-windows
        f.replace("{_zoned_datetime_str}", LogRecordNanoLocal.get_zoned_datetime_fmt())
        for f in [console_fmt, files_fmt]
    ]

    # logger
    logger_use = get_logger(logger=logger)  # skipif-ci-and-windows
    logger_use.setLevel(DEBUG)  # skipif-ci-and-windows

    # filters
    console_filters = (  # skipif-ci-and-windows
        [] if console_filters is None else list(console_filters)
    )
    files_filters = (  # skipif-ci-and-windows
        [] if files_filters is None else list(files_filters)
    )
    filters = (  # skipif-ci-and-windows
        [] if filters is None else list(always_iterable(filters))
    )

    # formatters
    try:  # skipif-ci-and-windows
        from coloredlogs import DEFAULT_FIELD_STYLES, ColoredFormatter
    except ModuleNotFoundError:  # pragma: no cover
        console_formatter = Formatter(fmt=console_fmt, style="{")
        files_formatter = Formatter(fmt=files_fmt, style="{")
    else:  # skipif-ci-and-windows
        field_styles = DEFAULT_FIELD_STYLES | {
            "_zoned_datetime_str": DEFAULT_FIELD_STYLES["asctime"]
        }
        console_formatter = ColoredFormatter(
            fmt=console_fmt, style="{", field_styles=field_styles
        )
        files_formatter = ColoredFormatter(
            fmt=files_fmt, style="{", field_styles=field_styles
        )
    plain_formatter = Formatter(fmt=files_fmt, style="{")  # skipif-ci-and-windows

    # console
    if console_level is not None:  # skipif-ci-and-windows
        console_low_handler = StreamHandler(stream=stdout)
        add_filters(console_low_handler, lambda x: x.levelno < ERROR)
        add_filters(console_low_handler, *console_filters)
        add_filters(console_low_handler, *filters)
        console_low_handler.setFormatter(console_formatter)
        console_low_handler.setLevel(console_level)
        logger_use.addHandler(console_low_handler)

        console_high_handler = StreamHandler(stream=stdout)
        add_filters(console_high_handler, *console_filters)
        add_filters(console_high_handler, *filters)
        _ = RichTracebackFormatter.create_and_set(
            console_high_handler,
            version=formatter_version,
            max_width=formatter_max_width,
            indent_size=formatter_indent_size,
            max_length=formatter_max_length,
            max_string=formatter_max_string,
            max_depth=formatter_max_depth,
            expand_all=formatter_expand_all,
            detail=True,
            post=_ansi_wrap_red,
        )
        console_high_handler.setLevel(
            max(get_logging_level_number(console_level), ERROR)
        )
        logger_use.addHandler(console_high_handler)

    # debug & info
    directory = resolve_path(path=files_dir)  # skipif-ci-and-windows
    levels: list[LogLevel] = ["DEBUG", "INFO"]  # skipif-ci-and-windows
    for level, (subpath, files_or_plain_formatter) in product(  # skipif-ci-and-windows
        levels, [(Path(), files_formatter), (Path("plain"), plain_formatter)]
    ):
        path = ensure_suffix(directory.joinpath(subpath, level.lower()), ".txt")
        path.parent.mkdir(parents=True, exist_ok=True)
        file_handler = SizeAndTimeRotatingFileHandler(
            filename=path,
            when=files_when,
            interval=files_interval,
            backupCount=files_backup_count,
            maxBytes=files_max_bytes,
        )
        add_filters(file_handler, *files_filters)
        add_filters(file_handler, *filters)
        file_handler.setFormatter(files_or_plain_formatter)
        file_handler.setLevel(level)
        logger_use.addHandler(file_handler)

    # errors
    standalone_file_handler = StandaloneFileHandler(  # skipif-ci-and-windows
        level=ERROR, path=directory.joinpath("errors")
    )
    add_filters(standalone_file_handler, lambda x: x.exc_info is not None)
    standalone_file_handler.setFormatter(
        RichTracebackFormatter(
            version=formatter_version,
            max_width=formatter_max_width,
            indent_size=formatter_indent_size,
            max_length=formatter_max_length,
            max_string=formatter_max_string,
            max_depth=formatter_max_depth,
            expand_all=formatter_expand_all,
            detail=True,
        )
    )
    logger_use.addHandler(standalone_file_handler)  # skipif-ci-and-windows

    # extra
    if extra is not None:  # skipif-ci-and-windows
        extra(logger_use)


##


@contextmanager
def temp_handler(
    handler: Handler, /, *, logger: LoggerOrName | None = None
) -> Iterator[None]:
    """Context manager with temporary handler set."""
    logger_use = get_logger(logger=logger)
    logger_use.addHandler(handler)
    try:
        yield
    finally:
        _ = logger_use.removeHandler(handler)


##


@contextmanager
def temp_logger(
    logger: LoggerOrName,
    /,
    *,
    disabled: bool | None = None,
    level: LogLevel | None = None,
    propagate: bool | None = None,
) -> Iterator[Logger]:
    """Context manager with temporary logger settings."""
    logger_use = get_logger(logger=logger)
    init_disabled = logger_use.disabled
    init_level = logger_use.level
    init_propagate = logger_use.propagate
    if disabled is not None:
        logger_use.disabled = disabled
    if level is not None:
        logger_use.setLevel(level)
    if propagate is not None:
        logger_use.propagate = propagate
    try:
        yield logger_use
    finally:
        if disabled is not None:
            logger_use.disabled = init_disabled
        if level is not None:
            logger_use.setLevel(init_level)
        if propagate is not None:
            logger_use.propagate = init_propagate


##


class _AdvancedLogRecord(LogRecord):
    """Advanced log record."""

    time_zone: ClassVar[str] = NotImplemented

    @override
    def __init__(
        self,
        name: str,
        level: int,
        pathname: str,
        lineno: int,
        msg: object,
        args: Any,
        exc_info: Any,
        func: str | None = None,
        sinfo: str | None = None,
    ) -> None:
        self._zoned_datetime = self.get_now()  # skipif-ci-and-windows
        self._zoned_datetime_str = (  # skipif-ci-and-windows
            self._zoned_datetime.format_common_iso()
        )
        super().__init__(  # skipif-ci-and-windows
            name, level, pathname, lineno, msg, args, exc_info, func, sinfo
        )

    @override
    def __init_subclass__(cls, *, time_zone: ZoneInfo, **kwargs: Any) -> None:
        cls.time_zone = time_zone.key  # skipif-ci-and-windows
        super().__init_subclass__(**kwargs)  # skipif-ci-and-windows

    @override
    def getMessage(self) -> str:
        """Return the message for this LogRecord."""
        msg = str(self.msg)  # pragma: no cover
        if self.args:  # pragma: no cover
            try:
                return msg % self.args  # compability for 3rd party code
            except ValueError as error:
                if len(error.args) == 0:
                    raise
                first = error.args[0]
                if search("unsupported format character", first):
                    return msg.format(*self.args)
                raise
            except TypeError as error:
                if len(error.args) == 0:
                    raise
                first = error.args[0]
                if search("not all arguments converted", first):
                    return msg.format(*self.args)
                raise
        return msg  # pragma: no cover

    @classmethod
    def get_now(cls) -> Any:
        """Get the current zoned datetime."""
        return cast("Any", ZonedDateTime).now(cls.time_zone)  # skipif-ci-and-windows

    @classmethod
    def get_zoned_datetime_fmt(cls) -> str:
        """Get the zoned datetime format string."""
        length = len(cls.get_now().format_common_iso())  # skipif-ci-and-windows
        return f"{{_zoned_datetime_str:{length}}}"  # skipif-ci-and-windows


##


def _ansi_wrap_red(text: str, /) -> str:
    try:
        from humanfriendly.terminal import ansi_wrap
    except ModuleNotFoundError:  # pragma: no cover
        return text
    return ansi_wrap(text, color="red")


__all__ = [
    "FilterForKeyError",
    "GetLoggingLevelNumberError",
    "SizeAndTimeRotatingFileHandler",
    "StandaloneFileHandler",
    "add_filters",
    "basic_config",
    "filter_for_key",
    "get_default_logging_path",
    "get_logger",
    "get_logging_level_number",
    "setup_logging",
    "temp_handler",
    "temp_logger",
]



from __future__ import annotations

from typing import Any

from hypothesis import given
from hypothesis.strategies import sampled_from

from utilities.reprlib import get_repr, get_repr_and_class


class TestGetRepr:
    @given(
        case=sampled_from([
            (None, "None"),
            (0, "0"),
            (
                list(range(21)),
                "[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, ... +1]",
            ),
        ])
    )
    def test_main(self, *, case: tuple[Any, str]) -> None:
        obj, expected = case
        result = get_repr(obj)
        assert result == expected


class TestGetReprAndClass:
    @given(
        case=sampled_from([
            (None, "Object 'None' of type 'NoneType'"),
            (0, "Object '0' of type 'int'"),
        ])
    )
    def test_main(self, *, case: tuple[Any, str]) -> None:
        obj, expected = case
        result = get_repr_and_class(obj)
        assert result == expected



from __future__ import annotations

import logging
from dataclasses import dataclass
from enum import StrEnum, unique
from logging import Handler, LogRecord
from sys import _getframe
from typing import TYPE_CHECKING, cast, override

from loguru import logger

from utilities.iterables import OneEmptyError, OneNonUniqueError, one

if TYPE_CHECKING:
    from collections.abc import Sequence

    from loguru import LevelConfig


LEVEL_CONFIGS: Sequence[LevelConfig] = [
    {"name": "TRACE", "color": "<blue><bold>"},
    {"name": "DEBUG", "color": "<cyan><bold>"},
    {"name": "INFO", "color": "<green><bold>"},
    {"name": "SUCCESS", "color": "<magenta><bold>"},
    {"name": "WARNING", "color": "<yellow><bold>"},
    {"name": "ERROR", "color": "<red><bold>"},
    {"name": "CRITICAL", "color": "<red><bold><blink>"},
]


##


class InterceptHandler(Handler):
    """Handler for intercepting standard logging messages.

    https://github.com/Delgan/loguru#entirely-compatible-with-standard-logging
    """

    @override
    def emit(self, record: LogRecord) -> None:
        # Get corresponding Loguru level if it exists.
        try:  # pragma: no cover
            level = logger.level(record.levelname).name
        except ValueError:  # pragma: no cover
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = _getframe(6), 6  # pragma: no cover
        while (  # pragma: no cover
            frame and frame.f_code.co_filename == logging.__file__
        ):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(  # pragma: no cover
            level, record.getMessage()
        )


##


@unique
class LogLevel(StrEnum):
    """An enumeration of the logging levels."""

    TRACE = "TRACE"
    DEBUG = "DEBUG"
    INFO = "INFO"
    SUCCESS = "SUCCESS"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


##


def get_logging_level_name(level: int, /) -> str:
    """Get the logging level name."""
    core = logger._core  # noqa: SLF001 # pyright: ignore[reportAttributeAccessIssue]
    try:
        return one(k for k, v in core.levels.items() if v.no == level)
    except OneEmptyError:
        raise _GetLoggingLevelNameEmptyError(level=level) from None
    except OneNonUniqueError as error:
        error = cast("OneNonUniqueError[str]", error)
        raise _GetLoggingLevelNameNonUniqueError(
            level=level, first=error.first, second=error.second
        ) from None


@dataclass(kw_only=True, slots=True)
class GetLoggingLevelNameError(Exception):
    level: int


@dataclass(kw_only=True, slots=True)
class _GetLoggingLevelNameEmptyError(GetLoggingLevelNameError):
    @override
    def __str__(self) -> str:
        return f"There is no level with severity {self.level}"


@dataclass(kw_only=True, slots=True)
class _GetLoggingLevelNameNonUniqueError(GetLoggingLevelNameError):
    first: str
    second: str

    @override
    def __str__(self) -> str:
        return f"There must be exactly one level with severity {self.level}; got {self.first!r}, {self.second!r} and perhaps more"


##


def get_logging_level_number(level: str, /) -> int:
    """Get the logging level number."""
    try:
        return logger.level(level).no
    except ValueError:
        raise GetLoggingLevelNumberError(level=level) from None


@dataclass(kw_only=True, slots=True)
class GetLoggingLevelNumberError(Exception):
    level: str

    @override
    def __str__(self) -> str:
        return f"Invalid logging level: {self.level!r}"


__all__ = [
    "LEVEL_CONFIGS",
    "GetLoggingLevelNameError",
    "GetLoggingLevelNumberError",
    "InterceptHandler",
    "LogLevel",
    "get_logging_level_name",
    "get_logging_level_number",
]



from __future__ import annotations

from asyncio import CancelledError, run, sleep
from contextlib import suppress
from logging import getLogger
from typing import override

from utilities.asyncio import AsyncService
from utilities.logging import basic_config
from utilities.random import SYSTEM_RANDOM

_LOGGER = getLogger(__name__)


class Service(AsyncService):
    @override
    async def _start(self) -> None:
        _LOGGER.info("Starting service...")

        async def coroutine() -> None:
            for i in range(1, 6):
                _LOGGER.info("Run #%d...", i)
                await sleep(0.1 + 0.4 * SYSTEM_RANDOM.random())
            _LOGGER.info("Cancelling...")
            raise CancelledError

        await coroutine()

    @override
    async def stop(self) -> None:
        _LOGGER.info("Stopping service...")
        await super().stop()


def main() -> None:
    basic_config()
    _LOGGER.info("Running script...")
    with suppress(CancelledError):
        run(_main())
    _LOGGER.info("Finished script")


async def _main() -> None:
    async with Service():
        ...



from __future__ import annotations

from typing import Any

from hypothesis import given

from tests.test_operator import make_objects
from utilities.hashlib import md5_hash


class TestMD5Hash:
    @given(obj=make_objects())
    def test_main(self, *, obj: Any) -> None:
        _ = md5_hash(obj)



from __future__ import annotations

import shutil
from contextlib import ExitStack, contextmanager
from dataclasses import dataclass
from pathlib import Path
from shutil import rmtree
from typing import TYPE_CHECKING, assert_never, override

from atomicwrites import replace_atomic

from utilities.errors import ImpossibleCaseError
from utilities.iterables import transpose
from utilities.tempfile import TemporaryDirectory

if TYPE_CHECKING:
    from collections.abc import Iterator

    from utilities.types import PathLike


def move(
    source: PathLike, destination: PathLike, /, *, overwrite: bool = False
) -> None:
    """Move/replace a file/directory atomically."""
    source, destination = map(Path, [source, destination])
    match (
        source.is_file(),
        source.is_dir(),
        destination.is_file(),
        destination.is_dir(),
        overwrite,
    ):
        case False, False, _, _, _:
            raise _MoveSourceNotFoundError(source=source)
        # files
        case (True, False, True, False, False) | (True, False, False, True, False):
            raise _MoveFileExistsError(source=source, destination=destination) from None
        case True, False, False, True, _:
            rmtree(destination, ignore_errors=True)
            return replace_atomic(str(source), str(destination))  # must be `str`s
        case True, False, _, _, _:
            return replace_atomic(str(source), str(destination))  # must be `str`s
        # directories
        case (False, True, True, False, False) | (False, True, False, True, False):
            raise _MoveDirectoryExistsError(source=source, destination=destination)
        case False, True, False, True, _:
            rmtree(destination, ignore_errors=True)
            _ = shutil.move(source, destination)
            return None
        case False, True, _, _, _:
            destination.unlink(missing_ok=True)
            _ = shutil.move(source, destination)
            return None
        case True, True, _, _, _:  # pragma: no cover
            raise ImpossibleCaseError(
                case=[f"{source.is_file()=}", f"{source.is_dir()=}"]
            )
        case _ as never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class MoveError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _MoveSourceNotFoundError(MoveError):
    source: Path

    @override
    def __str__(self) -> str:
        return f"Source {str(self.source)!r} does not exist"


@dataclass(kw_only=True, slots=True)
class _MoveFileExistsError(MoveError):
    source: Path
    destination: Path

    @override
    def __str__(self) -> str:
        return f"Cannot move file {str(self.source)!r} as destination {str(self.destination)!r} already exists"


@dataclass(kw_only=True, slots=True)
class _MoveDirectoryExistsError(MoveError):
    source: Path
    destination: Path

    @override
    def __str__(self) -> str:
        return f"Cannot move directory {str(self.source)!r} as destination {str(self.destination)!r} already exists"


##


def move_many(*paths: tuple[PathLike, PathLike], overwrite: bool = False) -> None:
    """Move a set of files concurrently."""
    sources, destinations = transpose(paths)
    with ExitStack() as stack:
        temp_paths = [
            stack.enter_context(writer(p, overwrite=overwrite)) for p in destinations
        ]
        for source, temp_path in zip(sources, temp_paths, strict=True):
            move(source, temp_path, overwrite=overwrite)


##


@contextmanager
def writer(path: PathLike, /, *, overwrite: bool = False) -> Iterator[Path]:
    """Yield a path for atomically writing files to disk."""
    path = Path(path)
    parent = path.parent
    parent.mkdir(parents=True, exist_ok=True)
    name = path.name
    with TemporaryDirectory(suffix=".tmp", prefix=name, dir=parent) as temp_dir:
        temp_path = Path(temp_dir, name)
        try:
            yield temp_path
        except KeyboardInterrupt:
            rmtree(temp_dir)
        else:
            try:
                move(temp_path, path, overwrite=overwrite)
            except _MoveSourceNotFoundError as error:
                raise _WriterTemporaryPathEmptyError(temp_path=error.source) from None
            except _MoveFileExistsError as error:
                raise _WriterFileExistsError(destination=error.destination) from None
            except _MoveDirectoryExistsError as error:
                raise _WriterDirectoryExistsError(
                    destination=error.destination
                ) from None


@dataclass(kw_only=True, slots=True)
class WriterError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _WriterTemporaryPathEmptyError(WriterError):
    temp_path: Path

    @override
    def __str__(self) -> str:
        return f"Temporary path {str(self.temp_path)!r} is empty"


@dataclass(kw_only=True, slots=True)
class _WriterFileExistsError(WriterError):
    destination: Path

    @override
    def __str__(self) -> str:
        return f"Cannot write to {str(self.destination)!r} as file already exists"


@dataclass(kw_only=True, slots=True)
class _WriterDirectoryExistsError(WriterError):
    destination: Path

    @override
    def __str__(self) -> str:
        return f"Cannot write to {str(self.destination)!r} as directory already exists"


__all__ = ["MoveError", "WriterError", "move", "move_many", "writer"]
